/**********************************************************************************************************************
 * Copyright (c) 2023 CenturyArks Co.,Ltd. All rights reserved.                                                       *
 *                                                                                                                    *
 * This source code or any portion thereof must not be  reproduced or used in any manner whatsoever.                  *
 * However, the extent permitted by the copyright holder is excluded.                                                 *
 **********************************************************************************************************************/
/* raw_mirror.c */
/* rawファイルミラー変換 Ver:0.50 2023-11-20版 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <time.h>

#define VECT_OUT_VAILD
#undef VECT_OUT_VAILD

#define HEADER_BUF_SIZE 2048
#define READ_DATA_BUF_SIZE 32768
#define WRITE_DATA_BUF_SIZE 8388608

#define MAX_WIDTH 3840

struct evtinfo {
    int height;
    int width;
    int evt_kind;   // 21:EVT2.1, 30:EVT3.0
    int endianness_legacy_flag;
    char header[HEADER_BUF_SIZE + 1];
    int header_len; // -1:無効
};

int get_evtinfo(struct evtinfo* eip, unsigned char* rbuf) {
    char * pos = strstr(rbuf, "\n% end\n");
    if(pos == NULL) {
        fprintf(stderr, "NG:Format file\n");
        eip->header_len = -1;
        return -1;
    }
    eip->header_len = (pos - (char *) rbuf) + 7;
    for(int i=0; i<eip->header_len; i++) {
        eip->header[i] = rbuf[i];
    }
    eip->header[eip->header_len] = '\0';
    pos = strstr(eip->header, "% format EVT");
    if(pos == NULL) {
        return -1;
    }
    char * pos2 = pos + 12;
    if(*(pos2) == '3' && *(pos2 + 1) == ';') {
        eip->evt_kind = 30;
    } else if(*(pos2) == '2' && *(pos2 + 1) == '1' && *(pos2 + 2) == ';') {
        eip->evt_kind = 21;
    } else {
        eip->evt_kind = 0;
    }
    pos = strstr(pos2, "endianness=legacy;");
    if(pos != NULL) {
        eip->endianness_legacy_flag = 1;
    } else {
        eip->endianness_legacy_flag = 0;
    }
    pos = strstr(pos2, "height=");
    if(pos == NULL) {
        return -1;
    }
    int tmpi = 0;
    for(int i=7; i<15; i++) {
        if((*(pos + i) == ';') || (*(pos + i) == '\n')) {
            break;
        }
        if((*(pos + i) >= '0') || (*(pos + i) <= '9')) {
            tmpi *= 10;
            tmpi += (*(pos + i) - '0');
        }
    }
    eip->height = tmpi;

    pos = strstr(pos2, "width=");
    if(pos == NULL) {
        return -1;
    }
    tmpi = 0;
    for(int i=6; i<14; i++) {
        if((*(pos + i) == ';') || (*(pos + i) == '\n')) {
            break;
        }
        if((*(pos + i) >= '0') || (*(pos + i) <= '9')) {
            tmpi *= 10;
            tmpi += (*(pos + i) - '0');
        }
    }
    eip->width = tmpi;

    if((eip->height == 0) || (eip->width == 0)) {
        return -1;
    }

    return eip->header_len;
}

void fout8(FILE* fp2, unsigned char uc1, unsigned char uc2, unsigned char uc3, unsigned char uc4,
                      unsigned char uc5, unsigned char uc6, unsigned char uc7, unsigned char uc8, int force_output) {
    static unsigned char tbuf[WRITE_DATA_BUF_SIZE];
    static int tbuf_pos = 0;

    if(force_output) {
        fwrite(tbuf, 1, tbuf_pos, fp2);
        return;
    }

    tbuf[tbuf_pos++] = uc1;
    tbuf[tbuf_pos++] = uc2;
    tbuf[tbuf_pos++] = uc3;
    tbuf[tbuf_pos++] = uc4;
    tbuf[tbuf_pos++] = uc5;
    tbuf[tbuf_pos++] = uc6;
    tbuf[tbuf_pos++] = uc7;
    tbuf[tbuf_pos++] = uc8;

    if(tbuf_pos >= WRITE_DATA_BUF_SIZE) {
        fwrite(tbuf, 1, tbuf_pos, fp2);
        tbuf_pos = 0;
    }
}

void fout2(FILE* fp2, unsigned char uc1, unsigned char uc2, int force_output) {
    static unsigned char tbuf[WRITE_DATA_BUF_SIZE];
    static int tbuf_pos = 0;

    if(force_output) {
        fwrite(tbuf, 1, tbuf_pos, fp2);
        return;
    }

    tbuf[tbuf_pos++] = uc1;
    tbuf[tbuf_pos++] = uc2;

    if(tbuf_pos >= WRITE_DATA_BUF_SIZE) {
        fwrite(tbuf, 1, tbuf_pos, fp2);
        tbuf_pos = 0;
    }
}

void vect_exec(FILE* fp2, char* x_buf, int width, int positive_flag) {
    int64_t out_buf = 0;
    int i2 = -1;
    unsigned char uc1, uc2;

    for(int i1 = width - 32; i1 >= 0; i1 -= 32) {
        out_buf = 0;

        for(int j = 0; j < 32; j++) {
            out_buf <<= 1;
            if(x_buf[i1 + j] == 1) {
                out_buf += 1;
            }
        }

        if(out_buf != 0) {
            // VECT 発行
            if(i2 != (width - i1 - 32)) {
                // VECT_BASE 必要
                uc2 = (positive_flag) ? 0x38 : 0x30;
                i2 = (width - i1 - 32);
                uc2 += ((i2 >> 8) & 0x07);
                uc1 = i2 & 0xff;
                fout2(fp2, uc1, uc2, 0);
            }
            // VECT12 #1
            uc2 = 0x40 + ((out_buf >> 8) & 0x0f);
            uc1 = out_buf & 0xff;
            fout2(fp2, uc1, uc2, 0);
            // VECT12 #2
            uc2 = 0x40 + ((out_buf >> 20) & 0x0f);
            uc1 = (out_buf >> 12) & 0xff;
            fout2(fp2, uc1, uc2, 0);
            // VECT8
            uc2 = 0x50;
            uc1 = (out_buf >> 24) & 0xff;
            fout2(fp2, uc1, uc2, 0);

            // Next Vect pos
            i2 += 32;
        }
    }
    for(int i=0; i < width; i++) { x_buf[i] = 0; }
}

unsigned char rev8(uint32_t v) {
    static unsigned char rev_table[256];
    static int first_flag = 1;
    if(first_flag) {
        for(int i = 0; i < 256; i++) {
            unsigned char uc1 = 0;
            for(int j = 0; j < 8; j++) {
                uc1 <<= 1;
                if((i >> j) & 0x01) {
                    uc1 += 1;
                }
            }
            rev_table[i] = uc1;
        }
        first_flag = 0;
    }
    return rev_table[(v & 0xff)];
}

void fout_EVT21_rev(FILE* fp2, int width, int endianness_legacy_flag, int64_t time_l, int y, int x, uint32_t valid, int positive_flag) {
    int adr_x = (width - 32) - x;   // 反転して、起点も反転する
    unsigned char uc8 = (positive_flag) ? 0x00 : 0x10;
    uc8 += ((time_l >> 2) & 0xff);
    unsigned char uc7 = ((time_l << 6) & 0xc0) + ((adr_x >> 5) & 0x3f);
    unsigned char uc6 = ((adr_x << 3) & 0xf8) + ((y >> 8) & 0x07);
    unsigned char uc5 = (y & 0xff);
    unsigned char uc4 = rev8(valid & 0xff);
    unsigned char uc3 = rev8((valid >> 8) & 0xff);
    unsigned char uc2 = rev8((valid >> 16) & 0xff);
    unsigned char uc1 = rev8((valid >> 24) & 0xff);
    if(endianness_legacy_flag) {
        fout8(fp2, uc5, uc6, uc7, uc8, uc1, uc2, uc3, uc4, 0);
    } else {
        fout8(fp2, uc1, uc2, uc3, uc4, uc5, uc6, uc7, uc8, 0);
    }
}

void evt21_mirror(FILE* fp2, int width, int endianness_legacy_flag, unsigned char* uc) {
    static int64_t time_h = 0;
    static int64_t time_l = 0;
    static int64_t tm1 = -1;

    int positive_flag = 1;

    int adr_y = 0;
    int adr_x = 0;

    unsigned char* uc_u;
    unsigned char* uc_l;
    if (endianness_legacy_flag) {
        uc_l = &uc[4];
        uc_u = uc;
    } else {
        uc_l = uc;
        uc_u = &uc[4];
    }

    int event = uc_u[3] >> 4;

    switch (event) {
    case 0:
        // EVT_NEG
        positive_flag = 0;
    case 1:
        // EVT_POS

        time_l = ((uc_u[3] << 2) & 0x3c) + ((uc_u[2] >> 6) & 0x03);

        adr_x = (uc_u[2] & 0x3f);
        adr_x = (adr_x << 5) + ((uc_u[1] >> 3) & 0x1f);

        adr_y = (uc_u[1] & 0x07);
        adr_y = (adr_y << 8) + uc_u[0];

        uint32_t valid = uc_l[3];
        valid = (valid << 8) + uc_l[2];
        valid = (valid << 8) + uc_l[1];
        valid = (valid << 8) + uc_l[0];
        fout_EVT21_rev(fp2, width, endianness_legacy_flag, time_l, adr_y, adr_x, valid, positive_flag);
        break;

    case 8:
        // EVT_TIME_HIGH
        time_h = uc_u[3] & 0x0f;
        time_h = (time_h << 8) + uc_u[2];
        time_h = (time_h << 8) + uc_u[1];
        time_h = (time_h << 8) + uc_u[0];
        time_h <<= 6;
        fout8(fp2, uc[0], uc[1], uc[2], uc[3], uc[4], uc[5], uc[6], uc[7], 0);
        break;

    case 10:
        // EVT_TRIGGER
        time_l = ((uc_u[3] & 0x0f) << 2) + ((uc_u[2] >> 6) & 0x03);
        fout8(fp2, uc[0], uc[1], uc[2], uc[3], uc[4], uc[5], uc[6], uc[7], 0);
        break;

    case 14:
        // OTHERS
        time_l = ((uc_u[3] & 0x0f) << 2) + ((uc_u[2] >> 6) & 0x03);
        fout8(fp2, uc[0], uc[1], uc[2], uc[3], uc[4], uc[5], uc[6], uc[7], 0);
        break;

    default:
        fout8(fp2, uc[0], uc[1], uc[2], uc[3], uc[4], uc[5], uc[6], uc[7], 0);
        break;
    }

    // 進行状況表示用
    if(tm1 != (int64_t) ((time_h + time_l) / 100000)) {
        tm1 = (int64_t) ((time_h + time_l) / 100000);
        fprintf(stderr, "Event Time:%.2f\n", (time_h + time_l) / 1000000.0);
    }
}


/**
 * @brief for EVT3.0 EVT_X_ADDR output
 * 
 * @param fp2 
 * @param x 
 * @param positive_flag 
 */
void fout_evt_x_addr(FILE* fp2, int x, int positive_flag) {
    unsigned uc1 = x & 0xff;
    unsigned uc2 = (positive_flag) ? 0x28 : 0x20;
    uc2 += ((x >> 8) & 0x07);
    fout2(fp2, uc1, uc2, 0);
}


void evt30_mirror(FILE* fp2, int width, unsigned char uc1, unsigned char uc2) {
    static int time_h = 0;
    static int time_l = 0;
    static int adr_y = 0;
    static int vect_base_x = 0;
    static int positive_flag = 1;
    
    static int disp_tm = 0;
    static int first_flag = 1;

    int adr_x = 0;
    int val1, pos;

#ifdef VECT_OUT_VAILD
    static char x_buf[MAX_WIDTH];
    if(first_flag) {
        // 念のためクリア
        for(int i = 0; i < MAX_WIDTH; i++) {
            x_buf[i] = 0;
        }
        first_flag = 0;
    }
#endif
    int event = (uc2 >> 4) & 0x0f;

    switch (event) {
    case 0: // EVT_ADDR_Y
        adr_y = uc2 & 0x07;
        adr_y = (adr_y << 8) + uc1;
#ifdef VECT_OUT_VAILD
        // 保持した x_buf を出力する処理
        vect_exec(fp2, x_buf, width, positive_flag);
#endif
        fout2(fp2, uc1, uc2, 0);

        break;

    case 2: // EVT_ADDR_X
        positive_flag = ((uc2 & 0x08) == 0x08) ? 1 : 0;
        adr_x = uc2 & 0x07;
        adr_x = (adr_x << 8) + uc1;
        vect_base_x = adr_x & 0xffe0;   // vect を更新するらしい下位５ビットはマスクする

        fout_evt_x_addr(fp2, (width - 1) - adr_x, positive_flag);
        break;

    case 3: // VECT_BASE_X
        positive_flag = ((uc2 & 0x08) == 0x08) ? 1 : 0;
        vect_base_x = uc2 & 0x07;
        vect_base_x = (vect_base_x << 8) + uc1;
        break;

    case 4: // VECT_12
    case 5: // VECT_8
        val1 = (event == 4) ? 12 : 8;
        pos = uc2 & 0x0f;
        pos = (pos << 8) + uc1;
        for (int x = 0; x < val1; x++) {
            if ((pos >> x) & 0x1) {
#ifdef VECT_OUT_VAILD
                // x_buf に設定する
                x_buf[vect_base_x + x] = 1;
#else
                fout_evt_x_addr(fp2, (width - 1) - (vect_base_x + x), positive_flag);
#endif
            }
        }
        vect_base_x += val1;
        break;

    case 6: // TIME_LOW
        time_l = uc2 & 0x0F;
        time_l = (time_l << 8) + uc1;
        // 時刻が変わるので
        // 保持した x_buf を出力する処理
#ifdef VECT_OUT_VAILD
        vect_exec(fp2, x_buf, width, positive_flag);
#endif
        fout2(fp2, uc1, uc2, 0);
        break;

    case 8: // TIME_HIGH
        time_h = uc2 & 0x0F;
        time_h = (time_h << 8) + uc1;
        time_h <<= 12;
        fout2(fp2, uc1, uc2, 0);
        // 進行状況表示用
        if(disp_tm != (int) (time_h / 100000)) {
            fprintf(stderr, "Event Time:%.2f\n", time_h / 1000000.0);
            disp_tm = (int) (time_h / 100000);
        }
        break;

    case 10: // EXT_TRIGGER
    case 14: // OTHERS
    case 15: // CONTINUED_12
    case 7:  // CONTINUED_4
    default: // 定義なし
        // そのまま送り出す
        fout2(fp2, uc1, uc2, 0);
        break;
    }
}


int main(int argc, char** argv) {
    unsigned char rx_data[READ_DATA_BUF_SIZE];
    FILE* fp1;
    FILE* fp2;

    clock_t startClk, endClk;
    startClk = clock();

    if(argc > 1) {
        char * file_name = argv[1];
        int flen = strlen(file_name);
        if(flen > 200) {
            fprintf(stderr, "Sorry file path is too long");
            exit(1);
        }
        if(flen < 4) {
            fprintf(stderr, "\nHow to use.\n\n%s [raw_file_name]\ninput [raw_file_name], output [raw_file_mame]_mirror.raw\n\nif no parameter then input stdin, output stdout\n", argv[0]);
            exit(0);     
        }
        if( ((file_name[flen - 1] == 'w') || (file_name[flen - 1] == 'W'))
         && ((file_name[flen - 2] == 'a') || (file_name[flen - 2] == 'A'))
         && ((file_name[flen - 3] == 'r') || (file_name[flen - 3] == 'R'))
         &&  (file_name[flen - 4] == '.') ) {
            file_name[flen - 4] = '\0';
         }
         
        char read_file_name[220];
        char write_file_name[220];
        sprintf(read_file_name, "%s.raw", file_name);
        sprintf(write_file_name, "%s_mirror.raw", file_name);


        fprintf(stderr, "SRC RAW:%s\nDST RAW:%s\n", read_file_name, write_file_name);
        fp1 = fopen(read_file_name, "rb");
        if(fp1 == NULL) {
            // 念のため不正文字列出力防止で、read_file_name は、表示しない
            fprintf(stderr, "Read file open fail.");
            exit(1);
        }
        fp2 = fopen(write_file_name, "wb");
        if(fp1 == NULL) {
            // READファイルは開けているので、不正文字はない（たぶん）から表示しとく
            fprintf(stderr, "Write file open fail : %s", write_file_name);
            exit(1);
        }
    } else {
        // 引数が無い場合は、標準入力で受けて、標準出力に出力する
        fp1 = stdin;
        fp2 = stdout;
    }

    char rbuf[HEADER_BUF_SIZE];
    int cnt = fread(rbuf, sizeof(char), HEADER_BUF_SIZE, fp1);

    struct evtinfo ei;

    int len1 = get_evtinfo(&ei, rbuf);
    if(len1 < 0) {
        fclose(fp1);
        fclose(fp2);
        fprintf(stderr, "Sorry could not parse header part");
        exit(1);
    }
    fprintf(stderr, "Header Length: %d\n", ei.header_len);
    fprintf(stderr, "-----\n%s-----\n", ei.header);
    fprintf(stderr, "height:%d, width:%d, EVT:%d, endianness_legacy_flag:%d\n", ei.height, ei.width, ei.evt_kind, ei.endianness_legacy_flag);
    if((ei.evt_kind != 30) && (ei.evt_kind != 21)) {
        fprintf(stderr, "Sorry current version support only EVT3.0 and EVT2.1");
        fclose(fp1);
        fclose(fp2);
        exit(1);
    }
    // データ開始位置に fseek する
    fseek(fp1, ei.header_len, 0);

    // ヘッダ部を出力
    fprintf(fp2, "%s", ei.header);

    while(1) {
        int rcnt = fread(rbuf, 1, READ_DATA_BUF_SIZE, fp1);
        if(ei.evt_kind == 30) {
            for(int i = 0; i < rcnt; i += 2) {
                evt30_mirror(fp2, ei.width, rbuf[i], rbuf[i + 1]);
            }
        } else {
            // EVT21
            for(int i = 0; i < rcnt; i += 8) {
                evt21_mirror(fp2, ei.width, ei.endianness_legacy_flag, &rbuf[i]);
            }
        }
        if(rcnt != READ_DATA_BUF_SIZE) {
            break;
        }
    }
    if(ei.evt_kind == 30) {
        fout2(fp2, 0, 0, 1);
    } else {
        fout8(fp2, 0, 0, 0, 0, 0, 0, 0, 0, 1);
    }

    fclose(fp1);
    fclose(fp2);

    endClk = clock();
    float cpusec = (float) (endClk - startClk) / CLOCKS_PER_SEC;
    fprintf(stderr, "CPU Time:%.2f\n", cpusec);   

    exit(0);

}
